/* Handle heaps (where a[k] <= a[2*k+1] && a[k] <= a[2*k+2]).
   Copyright (C) 1995 Free Software Foundation, Inc.
   Franois Pinard <pinard@iro.umontreal.ca>, 1995.
*/

/* Heaps are arrays for which a[k] <= a[2*k+1] and a[k] <= a[2*k+2]
   simultaneously, for all k, counting elements from 0.  For the sake of
   comparison, unexisting elements are considered to be infinite.  The
   interesting property of a heap is that a[0] is its smallest element.  */

#include "heap.h"

#ifndef NULL
# define NULL (void *) 0
#endif

extern void *xrealloc (void *, int);

/*-------------------------------------------------------------------.
| Initialize a HEAP structure, to be ordered by a COMPARE function.  |
`-------------------------------------------------------------------*/

void
init_heap (struct heap *heap, int (*compare) (void *, void *))
{
  heap->compare = compare;
  heap->array = NULL;
  heap->size = 0;
  heap->allocated = 0;
}

/*-----------------------------.
| Terminate a HEAP structure.  |
`-----------------------------*/

void
term_heap (struct heap *heap)
{
  if (heap->array)
    free (heap->array);
}

/*----------------------------------------------------------------------.
| Resift a[0] of a HEAP structure, for when it just has been replaced.  |
`----------------------------------------------------------------------*/

void
resift_heap (struct heap *heap)
{
  int low;
  int high;
  void *work;

  for (low = 0, high = 1; high < heap->size; low = high, high = 2*low + 1)
    {
      if (high + 1 < heap->size
	  && (*heap->compare) (heap->array[high], heap->array[high + 1]) > 0)
	high++;

      if ((*heap->compare) (heap->array[low], heap->array[high]) > 0)
	{
	  work = heap->array[low];
	  heap->array[low] = heap->array[high];
	  heap->array[high] = work;
	}
      else
	break;
    }      
}

/*--------------------------------------.
| To a HEAP structure, add a new ITEM.  |
`--------------------------------------*/

void
increase_heap (struct heap *heap, void *item)
{
  int low;
  int high;
  void *work;

  if (heap->size == heap->allocated)
    {
      heap->allocated += 200;
      heap->array = xrealloc (heap->array, heap->allocated * sizeof (void *));
    }
  heap->array[heap->size++] = item;

  for (high = heap->size - 1; high > 0; high = low)
    {
      low = (high - 1) / 2;
      if ((*heap->compare) (heap->array[low], heap->array[high]) > 0)
	{
	  work = heap->array[low];
	  heap->array[low] = heap->array[high];
	  heap->array[high] = work;
	}
      else
	break;
    }
}

/*-------------------------------------------------.
| From a HEAP structure, delete the zero'th item.  |
`-------------------------------------------------*/

void
decrease_heap (struct heap *heap)
{
  heap->array[0] = heap->array[--heap->size];
  resift_heap (heap);
}

/* Trying it.  */

#if TEST

const char *program_name;

int
comparer (void *a, void *b)
{
  return *(int *) a - *(int *) b;
}

main (int argc, char *const *argv)
{
  int data[100];
  int k;
  struct heap h;

  program_name = argv[0];

  for (k = 0; k < 100; k++)
    data[k] = 9000+k;

  init_heap (&h, comparer);
  for (k = 100 - 1; k >= 0; k--)
    increase_heap (&h, data + k);

  for (k = 0; k < 100; k++)
    printf (" %d-%d", k, *(int *) h.array[k]);
  printf ("\n\n\n");

  while (h.size > 0)
    {
      printf (" %d-%d", 100-h.size, *(int *) h.array[0]);
      decrease_heap (&h);
    }

  exit (0);
}
#endif
